{% extends "!footer.html" %}
{% block extrafooter %}
    <p>The design of this documentation has been inspired by the <a href="http://scanpy.readthedocs.io">documentation</a> of the Python package <a href="https://github.com/theislab/scanpy">scanpy</a></p>
    {{ super() }}
{% endblock %}
    
